''' Serializers/Deserializers (SerDe) convert between configuration strings (or numbers) and
different Python structures.

You should subclass :class:`ConfigSerDeBase` and override the following static fields:

- ``name``: the description of the structure this serde can convert, for example "Comma-separated
  list of integers". This text will be used in the documentation generated by the ConfigBase CLI.
- ``serialize``: Take the Python structure and returns a string representation.
- ``deserialize``: Take a string representation of the value and return a Python structure.
'''

from typing import Any, List, Sequence, Mapping
from abc import ABC, abstractmethod
import json
import datetime

class ConfigSerDeBase(ABC):
    ''' Defines a serializer / deserializer interface. '''

    description : str = 'The base serde encodes and decodes strings to themselves.'
    ''' A textual description how the object is encoded in the string. Will be used in docs. '''

    example : str = 'abcdefgh'
    ''' Provide a textual example of the encoded string. Will be used in docs. '''

    @staticmethod
    @abstractmethod
    def serialize(value: Any, metadata: Mapping[str, Any]={}) -> str:
        ''' Serializes a config value to a string.

        Args:
            value (Any): a Python object to be serialized.
            metadata (Mapping[str, Any]): Additional metadata to be passed to SerDe implementations.

        Returns:
            The object serialized into a string.
        '''
        return str(value)

    @staticmethod
    @abstractmethod
    def deserialize(value: str, metadata: Mapping[str, Any]={}) -> Any:
        ''' Deserializes a string to a config value.

        Args:
            value (str): a Python object serialized into a string
            metadata (Mapping[str, Any]): Additional metadata to be passed to SerDe implementations.

        Returns:
            The the restored Python object.
        '''
        return value


class IntegerListSerDe(ConfigSerDeBase):
    ''' De/serializes a string containing a comma-separated list of integers.'''

    description : str = 'Comma-separated list of integers'
    example: str = '0, 1, 2'

    @staticmethod
    def serialize(value: Sequence[int], metadata: Mapping[str, Any]={}) -> str:
        ''' Serializes a list of integers into a string.

        Args:
            value (Sequence[int]): The list of integers.

        Returns:
            The list of integers serialized into a string.
        '''
        sep = metadata.get('separator', ', ')
        return sep.join(str(e) for e in value)

    @staticmethod
    def deserialize(value: str, metadata: Mapping[str, Any]={}) -> List[int]:
        '''Restores a list of integers from a string.

        Args:
            value (str): A string containing a serialized list of integers.

        Returns:
            The list of integers restored from the string.

        Raises:
            Exception: exceptions related to invalid string format.
        '''
        sep = metadata.get('separator', ',')
        return [int(e) for e in value.split(sep)]


class StringListSerDe(ConfigSerDeBase):
    ''' De/serializes a string containing a comma-separated list of strings.'''

    description : str = 'Comma-separated list of strings'
    example: str = 'apple, pear, banana'

    @staticmethod
    def serialize(value: Sequence[str], metadata: Mapping[str, Any]={}) -> str:
        ''' Serializes a list of strings into a string.

        Args:
            value (Sequence[str]): The list of strings.

        Returns:
            The list of strings serialized into a string.
        '''
        sep = metadata.get('separator', ', ')
        return sep.join(str(e) for e in value)

    @staticmethod
    def deserialize(value: str, metadata: Mapping[str, Any]={}) -> List[str]:
        '''Restores a list of strings from a string.

        Args:
            value (str): A string containing a serialized list of strings.

        Returns:
            The list of strings restored from the string.

        Raises:
            Exception: exceptions related to invalid string format.
        '''
        sep = metadata.get('separator', ',')
        return [str(e).strip() for e in value.split(sep)]


class JsonSerDe(ConfigSerDeBase):
    ''' De/serializes a string containing a JSON-encoded object.'''

    description : str = 'JSON-encoded object'
    example: str = '{"foo": "bar", "list": [0, 1, 2]}'

    @staticmethod
    def serialize(value: Any, metadata: Mapping[str, Any]={}) -> str:
        ''' Serializes a JSON object into a string.

        Args:
            value (Any): A JSON-serializable object.

        Returns:
            The JSON object serialized into a string.
        '''
        return json.dumps(value)

    @staticmethod
    def deserialize(value: str, metadata: Mapping[str, Any]={}) -> Any:
        ''' Deserializes JSON object from a string.

        Args:
            value (str): A string containing a serialized JSON object.

        Returns:
            The JSON object restored from the string.

        Raises:
            Exception: exceptions related to invalid string format.
        '''
        return json.loads(value)


class TimeDeltaSecondsSerDe(ConfigSerDeBase):
    ''' De/serializes a timedelta instance into a float number.

    The time interval is expressed in seconds.
    '''

    name : str = 'float32 value interpreted as seconds'

    @staticmethod
    def serialize(value: datetime.timedelta, metadata: Mapping[str, Any]={}) -> float:
        ''' Serializes a timedelta instance as a float number.

        The time interval is expressed in seconds. '''
        return value.total_seconds()


    @staticmethod
    def deserialize(value: float, metadata: Mapping[str, Any]={}) -> datetime.timedelta:
        ''' Deserializes a float value into a time interval.

        The value is interpreted in seconds.
        '''
        return datetime.timedelta(seconds=value)



''' This module contains the :class:`AutoIdentity` class that provides information about the
application execution environment. '''

import os
import logging
import datetime
from typing import Dict, Optional, Iterator, Any
import time

import boto3
from pydantic import BaseModel, Field

class AutoIdentityData(BaseModel):
    ''' Data class to store auto identity information. '''

    application_instance_id: str = Field(alias='ApplicationInstanceId')
    ''' Application instance id. '''

    application_name: str = Field(alias='Name')
    ''' Name of this application. '''

    application_tags: Dict[str, str] = Field(alias='Tags')
    ''' Tags associated with this application. '''

    device_id: str = Field(alias='DefaultRuntimeContextDevice')
    ''' Device id of the appliance running this application. '''

    device_name: str = Field(alias='DefaultRuntimeContextDeviceName')
    '''  Name of this application. '''

    application_created_time: datetime.datetime = Field(alias='CreatedTime')
    ''' Application deployment time. '''

    application_status: str = Field(alias='HealthStatus')
    ''' Health status of this application. '''

    application_description: str = Field(alias='Description')
    ''' The description of this application. '''

    @classmethod
    def for_test_environment(cls, application_instance_id: str, application_name: str):
        ''' Initializes a dummy AutoIdentityData to be used in test environment. '''
        return cls(
            ApplicationInstanceId=application_instance_id,
            Name=application_name,
            Tags={},
            DefaultRuntimeContextDevice='emulator',
            DefaultRuntimeContextDeviceName='test_utility_emulator',
            CreatedTime=datetime.datetime.now(),
            HealthStatus='TEST_UTILITY',
            Description=application_name,
        )


class AutoIdentityFetcher:
    ''' AutoIdentity instance queries metadata of the current application instance.

    The IAM policy associated with the `Panorama Application Role`_
    of this app should grant the execution of
    `panorama:ListApplicationInstances`_ operation.

    Args:
        device_region: The AWS region where this Panorama appliance is registered.
        application_instance_id: The application instance id. If left to `None`,
            :class:`AutoIdentity` will try to find the instance id in the environment variable.
        parent_logger: If you want to connect the logger to a parent, specify it here.

    .. _`Panorama Application Role`:
        https://docs.aws.amazon.com/panorama/latest/dev/permissions-application.html
    .. _`panorama:ListApplicationInstances`:
        https://docs.aws.amazon.com/service-authorization/latest/reference/list_awspanorama.html#awspanorama-actions-as-permissions
    '''

    # pylint: disable=too-many-instance-attributes,too-few-public-methods
    # This class functions as a data class that reads its values from the environment

    def __init__(self,
        device_region: str,
        application_instance_id: Optional[str] = None,
        parent_logger: Optional[logging.Logger] = None
    ):
        self._logger = (
            logging.getLogger(self.__class__.__name__) if parent_logger is None else
            parent_logger.getChild(self.__class__.__name__)
        )
        self.application_instance_id = (
            application_instance_id or os.environ.get('AppGraph_Uid')
        )
        if not self.application_instance_id:
            raise RuntimeError(
                'Could not find application instance id in environment variable "AppGraph_Uid"'
            )
        self.device_region = device_region

    def get_data(self, retry_freq: Optional[float] = None) -> AutoIdentityData:
        ''' Fetches the auto identity data.

        Args:
            retry_freq (Optional[float]): If set to a float number, AutoIdentity will keep retrying
                fetching the auto identity data from remote services if the status of the app was
                "NOT_AVAILABLE". If set to None, will not retry.

        Raises:
            RuntimeError: if could not fetch the auto identity information, and retry_freq is set
                to None.
        '''

        def fetch() -> Dict[str, Any]:
            app_instance_data = self._app_instance_data(self.application_instance_id)
            if not app_instance_data:
                raise RuntimeError(
                    'Could not find application instance in service response. '
                    'Check if application_instance_id=%s '
                    'and device_region=%s parameters are correct.'
                    .format(self.application_instance_id, self.device_region)
                )
            else:
                return app_instance_data

        while True:
            app_instance_data = fetch()
            status = app_instance_data.get('HealthStatus', 'NOT_AVAILABLE')
            if status == 'NOT_AVAILABLE':
                if retry_freq is None:
                    raise RuntimeError(
                        'Application HealthStatus is "NOT_AVAILABLE" and retry is disabled.'
                    )
                else:
                    time.sleep(retry_freq)
                    continue
            elif status == 'ERROR':
                raise RuntimeError('Application HealthStatus is "ERROR"')
            else:
                return AutoIdentityData(**app_instance_data)

    def _list_app_instances(self, deployed_only=True) -> Iterator[Dict[str, Any]]:
        session = boto3.Session(region_name=self.device_region)
        panorama = session.client('panorama')
        next_token = None
        while True:
            kwargs = {'StatusFilter': 'DEPLOYMENT_SUCCEEDED'} if deployed_only else {}
            if next_token:
                kwargs['NextToken'] = next_token
            response = panorama.list_application_instances(**kwargs)
            inst: Dict[str, Any]
            for inst in response['ApplicationInstances']:
                yield inst
            if 'NextToken' in response:
                next_token = response['NextToken']
            else:
                break

    def _app_instance_data(self, application_instance_id) -> Optional[Dict[str, Any]]:
        matches = (inst
            for inst in self._list_app_instances()
            if inst.get('ApplicationInstanceId') == application_instance_id
        )
        return next(matches, None)



''' The ``config.tool`` module defines a simple command-line interface (CLI) program that can be
used as a development tool. It creates snippets that you can copy-paste into the Panorama
application project files like graph.json, or package.json of the business logic package.
'''
from typing import Mapping, Any
import sys
import argparse, json, textwrap
from .config import ConfigBase

def cli(name: str, config: ConfigBase) -> None:
    ''' Creates a simple command line interface for printing config snippets.

    Args:
        name (str): The name of your Panorama app
        config (ConfigBase): Instance of your configuration structure, subclass of ConfigBase.
    '''
    def render_template(template_path: str, variables: Mapping[str, Any]):
        try:
            from jinja2 import Environment, FileSystemLoader
        except ImportError:
            sys.exit('To use the render command you must install jinja2 with "pip install Jinja2"')
        env = Environment(loader=FileSystemLoader('/'), trim_blocks=True)
        env.filters['to_pretty_json'] = lambda value: json.dumps(value, indent=4)
        template = env.get_template(template_path)
        return template.render(variables)

    parser = argparse.ArgumentParser(description=textwrap.dedent(
        f'''\
        Configuration snippet generator for {name} application.

        This program can generate json and markdown snippets that you can copy-paste to the
        metadata and package definitions of your AWS Panorama project. The snippets contain the
        definitions of the application parameters in the required format. The following formats
        are supported:

         - nodes: generates a json snippet to be pasted in the nodeGraph.nodes field of graph.json
         - edges: generates a json snippet to be pasted in the nodeGraph.edges field of graph.json.
            Specify the code node name.
         - interface: generates json a snippet to be pasted in nodePackage.interfaces field of the
             package.json of the application code package
         - markdown: generates a markdown snippet that you can paste to the README of your project,
             or other parts of the documentation.
         - render: renders a Jinja2 template. Specify the template filename and the code node name.
         '''),
         formatter_class=argparse.RawDescriptionHelpFormatter
    )
    function_map = {
        'nodes' : lambda config, _: (
            print(json.dumps(config.get_panorama_definitions(), indent=4))
        ),
        'edges' : lambda config, kwargs: (
            print(json.dumps(config.get_panorama_edges(kwargs['code_node']), indent=4))
        ),
        'interface': lambda config, _: (
            print(json.dumps(config.get_panorama_app_interface(), indent=4))
        ),
        'markdown': (
            lambda config, _: print(config.get_panorama_markdown_doc())
        ),
        'render': lambda config, kwargs: (
            print(render_template(kwargs['template'], {
                'nodes': config.get_panorama_definitions(),
                'edges': config.get_panorama_edges(kwargs['code_node']),
                'interface': config.get_panorama_app_interface(),
                'markdown': config.get_panorama_markdown_doc()
            }))
        )
    }

    parser.add_argument('command', choices=function_map.keys(),
        help='Prints configuration snippets for graph.json nodes, edges, application '
             'interface in package.json, or in markdown format.'
    )
    parser.add_argument('--code-node', '-c', type=str, default='code_node',
        help='Code node name (used in edges snippet)'
    )
    parser.add_argument('--template', '-t', type=str,
        help='Template file (used in render command)'
    )

    args = parser.parse_args()
    func = function_map[args.command]
    func(config, {'code_node': args.code_node, 'template': args.template})

if __name__=='__main__':
    cli('config_base', ConfigBase())



''' Reports timer statistics to AWS CloudWatch Metrics. '''

import datetime
import logging
from typing import Optional, Dict

import boto3
import botocore

from .timepiece import BaseTimer

class CloudWatchTimerAdapter:
    ''' Reports timer statistics to AWS CloudWatch Metrics.

    The IAM policy associated with the Panorama Application Role of this app should grant
    the execution of `cloudwatch:PutMetricData` operation.

    Args:
        namespace (str): The name of the CloudWatch namespace of this custom metrics.
            It can be for example the name of your project.
        metric_name (str): The name of the CloudWatch metrics. This can be for example
            `frame_processing_time`, if you use CWTachometer to measure frame processing
            time statistics.
        dimensions (Optional[Dict[str, str]]): Additional CloudWatch metrics dimensions of this
            metric. This can be for example the device and application identifier.
        region (Optional[str]): The AWS region of the CloudWatch metrics.
        boto3_session (Optional[boto3.Session]): The boto3 session to be used for sending the
            CloudWatch metrics. If left to None, CWTachometer will use the default session. If the
            default session does not have a default region configured, you might get errors.
        parent_logger (Optional[logging.Logger]): If you want to connect the logger of this class
            to a parent, specify it here.
    '''

    def __init__(self,
        namespace: str,
        metric_name: str,
        dimensions: Optional[Dict[str, str]] = None,
        region: Optional[str] = None,
        boto3_session: Optional[boto3.Session] = None,
        parent_logger: Optional[logging.Logger] = None
    ):
        self.logger = (
            logging.getLogger(self.__class__.__name__) if parent_logger is None else
            parent_logger.getChild(self.__class__.__name__)
        )
        self.dimensions = CloudWatchTimerAdapter._cw_dimensions(dimensions or {})
        self.namespace = namespace
        self.metric_name = metric_name
        boto3_session = boto3_session or boto3.Session(region_name=region)
        self.cloudwatch = boto3_session.client('cloudwatch')

    @staticmethod
    def _cw_dimensions(dimensions):
        return [{ 'Name': name, 'Value': value } for name, value in dimensions.items()]

    def send_metrics(self, timestamp: datetime.datetime, timer: BaseTimer) -> None:
        ''' Sends timer statistics to CloudWatch.

        This method can be used as a callback in Tachometer instances.

        For example::

            cw_adapter = CloudWatchTimerAdapter(
                namespace='my_namespace',
                metric_name='my_metric',
                dimensions={'foo': 'bar'}
            )
            tacho = TickerTachometer(
                stats_callback=cw_adapter.send_metrics
            )
            tacho.tick()

        Args:
            timestamp (datetime.datetime): The timestamp the statistics refers to.
            timer (BaseTimer): The timer that collected the statistics.
        '''
        metric_data = {
            'MetricName': self.metric_name,
            'Dimensions': self.dimensions,
            'Timestamp': timestamp.astimezone(datetime.timezone.utc),
            'StatisticValues': {
                'SampleCount': timer.len(),
                'Sum': timer.sum(),
                'Minimum': timer.min(),
                'Maximum': timer.max()
            },
            'Unit': 'Seconds'
        }
        try:
            self.cloudwatch.put_metric_data(
                Namespace=self.namespace,
                MetricData=[metric_data]
            )
        except botocore.exceptions.ClientError as error:
            self.logger.warning('Couldn\'t put data for metric %s.%s',
                                self.namespace, self.metric_name)
            self.logger.warning(str(error))
        except AttributeError:
            self.logger.warning('CloudWatch client is not available.')
        return metric_data



''' This module defines :class:`~backpack.config.ConfigBase`, a base class for Panorama application
configurations. The class offers two basic functionalities:

- parse parameters from the Panorama application input ports (``panoramasdk.node.inputs``)
- generate configuration file snippets for ``graph.json`` and ``package.json`` in your Panorama
  project. For more information about how to use the CLI, refer to :meth:`~backpack.config.tool`.
'''

import dataclasses
from typing import Sequence, List, Any, Type, TypeVar, Mapping, Tuple, Optional
import textwrap

from .serde import ConfigSerDeBase

T = TypeVar('T', bound='ConfigBase')

class ConfigBase:
    ''' Base class for configuration structures.

    Subclasses must be also dataclasses.
    '''

    TYPE_MAP = {
        int: 'int32',
        float: 'float32',
        str: 'string',
        bool: 'boolean'
    }

    def __init__(self) -> None:
        assert dataclasses.is_dataclass(self), 'ConfigBase instances must be also dataclasses.'

    @staticmethod
    def _get_param_name(full_path: Sequence[str]=[]) -> str:
        return '_'.join(full_path)

    @staticmethod
    def _get_param_type(field: dataclasses.field) -> str:
        if 'type' in field.metadata:
            typename = field.metadata['type']
        else:
            typename = ConfigBase.TYPE_MAP.get(field.type)
        if typename is None:
            raise ValueError(f'Field has unsupported type: {field}')
        return typename

    @staticmethod
    def _get_param_serde(field: dataclasses.field) -> Optional[Type[ConfigSerDeBase]]:
        return field.metadata.get('serde')

    @staticmethod
    def _get_param_default(
        field: dataclasses.field,
        serde_metadata: Mapping[str, Any]
    ) -> Optional[str]:
        if field.default is not dataclasses.MISSING:
            default = field.default
        elif field.default_factory  is not dataclasses.MISSING:
            default = field.default_factory()
        else:
            default = None
        serde = ConfigBase._get_param_serde(field)
        if serde is not None and default is not None:
            default = serde.serialize(default, metadata=serde_metadata)
        return default

    @staticmethod
    def _get_param_doc(field: dataclasses.field) -> Optional[str]:
        doc: str = field.metadata.get('__doc__', field.metadata.get('doc'))
        doc = textwrap.dedent(doc)
        doc = doc.replace('\n', ' ')
        doc = doc.replace('  ', ' ')
        doc = doc.strip()
        return doc

    def _param_walker(self, _current_path: Sequence[str]=[]) -> Tuple[str, dataclasses.field]:
        ''' Recursively walks all parameters in the config structure.

        Args:
            _current_path (Sequence[str]): The current path in the config structure. This is an
                internal recursion parameter and users should always leave the default empty
                list value.

        Returns:
            A generator that yields a tuple for each parameter. The tuple consists of the
            following values:
                - full_path (Sequence[str]): The full path of the parameter in the hierarchy
                - field (dataclasses.field): The original field of the
        '''
        fields = dataclasses.fields(self)
        for fld in fields:
            current_path = _current_path + [fld.name]
            if dataclasses.is_dataclass(fld.type):
                obj = getattr(self, fld.name)
                for sub_result in obj._param_walker(_current_path=current_path):
                    yield sub_result
            else:
                yield (current_path, fld)

    def get_panorama_definitions(self, serde_metadata: Mapping[str, Any]={}) -> List[Mapping[str, Any]]:
        ''' Generate the ``nodeGraph.nodes`` snippet in ``graph.json``.

        Returns:
            A list of dictionaries containing the application parameter node definitions.
        '''
        return [
            {
                'name': ConfigBase._get_param_name(full_path=full_path),
                'interface': ConfigBase._get_param_type(field=field),
                'value': ConfigBase._get_param_default(field=field, serde_metadata=serde_metadata),
                'overridable': True,
                'decorator': {
                    'title': ConfigBase._get_param_name(full_path=full_path),
                    'description': ConfigBase._get_param_doc(field=field)
                }
            }
            for (full_path, field) in self._param_walker()
        ]

    def get_panorama_edges(self, code_node_name: str) -> List[Mapping[str, str]]:
        ''' Generate the ``nodeGraph.edges`` snippet in ``graph.json``

        Returns:
            A list of dictionaries containing the application edge definitions.
        '''
        return [
            {
                "producer": ConfigBase._get_param_name(full_path=full_path),
                "consumer": code_node_name + "." + ConfigBase._get_param_name(full_path=full_path)
            }
            for (full_path, _) in self._param_walker()
        ]

    def get_panorama_app_interface(self) -> List[Mapping[str, str]]:
        ''' Generate the application interface snippet in app node ``package.json``.

        Returns:
            A list of dictionaries containing the elements of the application interface definition.
        '''
        return [
            {
                "name": ConfigBase._get_param_name(full_path=full_path),
                "type": ConfigBase._get_param_type(field=field)
            }
            for (full_path, field) in self._param_walker()
        ]

    def get_panorama_markdown_doc(self, serde_metadata: Mapping[str, Any]={}) -> str:
        ''' Generates a markdown table of the parameters that can be used in documentation.

        Returns:
            Markdown formatted text containing the parameter documentation.
        '''
        header = (
            '| name | type    | default | description |\n'
            '|------|---------|---------|-------------|\n'
        )
        body = '\n'.join([
            f'| {ConfigBase._get_param_name(full_path=full_path)} '
            f'| {ConfigBase._get_param_type(field=field)} '
            f'| {ConfigBase._get_param_default(field=field, serde_metadata=serde_metadata)} '
            f'| {ConfigBase._get_param_doc(field=field)} |'
            for (full_path, field) in self._param_walker()
        ])
        return header + body

    @classmethod
    def from_panorama_params(cls: Type[T],
        inputs: 'panoramasdk.port', # type: ignore
        serde_metadata: Mapping[str, Any]={}
    ) -> T:
        ''' Parses the config values form AWS Panorama input parameters.

        A new Config object is created with the default values. If a particular value is
        found in the input parameter, its value will override the default value.

        Args:
            inputs (panoramasdk.port): The input port of the Panorama application node.

        Returns:
            The config instance filled with the parameter values read from the input port.
        '''
        result = cls()

        for (full_path, fld) in result._param_walker():
            obj = result
            for name_part in full_path[:-1]:
                obj = getattr(obj, name_part)
            key = full_path[-1]
            name = ConfigBase._get_param_name(full_path=full_path)
            if not hasattr(inputs, name):
                continue
            value = getattr(inputs, name).get()
            serde = ConfigBase._get_param_serde(field=fld)
            if serde is not None:
                value = serde.deserialize(value, metadata=serde_metadata)
            setattr(obj, key, value)
        return result

    def asdict(self):
        return dataclasses.asdict(self)



''' This module contains code to standardize the configuration of Panorama applications via
deploy-time parameters. It also contains a CLI (command-line interface) that generates
configuration snippets for the metadata and descriptor json files of your Panorama project.
'''

from .tool import cli
from .config import ConfigBase
from .serde import (
    ConfigSerDeBase, IntegerListSerDe, StringListSerDe, JsonSerDe, TimeDeltaSecondsSerDe
)



''' This module contains config serdes for objects from backpack.geometry module. '''

from typing import Optional, Mapping, Any
import dataclasses
import json

from backpack.geometry import PolyLine

from .serde import ConfigSerDeBase

class PolyLineSerDe(ConfigSerDeBase):
    ''' De/serializes a json string containing a polygon definition.

    Example string: ``[[0.1, 0.1], [0.9, 0.1], [0.9, 0.9], [0.5, 0.5], [0.1, 0.9]]``
    '''
    description : str = 'JSON encoded Polygon'
    example: str = '[[0.1, 0.1], [0.9, 0.1], [0.9, 0.9], [0.5, 0.5], [0.1, 0.9]]'

    @staticmethod
    def serialize(value: PolyLine, metadata: Mapping[str, Any]={}) -> str:
        ''' Serializes a PolyLine into a string.

        Args:
            value (PolyLine): The PolyLine to be serialized.

        Returns:
            The PolyLine serialized into a string.
        '''
        return json.dumps(dataclasses.astuple(value)[0])

    @staticmethod
    def deserialize(value: str, metadata: Mapping[str, Any]={}) -> Optional[PolyLine]:
        '''
        Restores a PolyLine from a string.

        Args:
            value (str): A string containing a serialized PolyLine.

        Returns:
            The PolyLine restored from the string or None if the string was empty.

        Raises:
            Exception: exceptions related to invalid string format.
        '''
        return (
            PolyLine.from_value(json.loads(value))
                if value is not None and len(value) > 0 else
            None
        )



import os
import codecs
import os.path

from setuptools import setup, find_packages

name = "panorama-backpack"

with open("requirements.txt", "r") as f:
    requires = [x.strip() for x in f if x.strip()]

with open("README.md", "r", encoding="utf-8") as f:
    long_description = f.read()

def read(rel_path):
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, rel_path), 'r') as fp:
        return fp.read()

def get_version(rel_path):
    for line in read(rel_path).splitlines():
        if line.startswith("__version__"):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    else:
        raise RuntimeError("Unable to find version string.")

version=get_version("backpack/__init__.py")

setup(
    name=name,
    version=version,
    description="Tools for AWS Panorama development",
    python_requires=">=3.7",
    packages=find_packages(exclude=["tests.*", "tests"]),
    package_dir={"backpack": "backpack"},
    include_package_data=True,
    install_requires=requires,
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Janos Tolgyesi",
    author_email='janos.tolgyesi@neosperience.com',
    url="https://github.com/Neosperience/backpack",
    download_url="https://github.com/Neosperience/backpack/archive/refs/tags/v{}.tar.gz".format(version),
    keywords=["aws", "panorama", "video-analysis"],
    entry_points={},
    zip_safe=False,
    classifiers=[
        "Programming Language :: Python :: 3.7",
    ],
    extras_require={
        "opencv": ["opencv-headless"]
    },
    test_suite='tests'
)
